package bc_packetint;

use bc_enc;
use bc_handleproto;
use bc_global;
use bc_modules;
use bc_command;

use strict;

# Pack Interpreter

use Switch;

sub event {
}

sub pi {
  my @packet = @_;

  if ($packet[0] eq "second") {
    event();
    return 0;
  }

  if ($packet[0] eq "undef_p") {
    our (@p, @ids);
    undef @p;
    undef @ids;
    return 0;
  }

  switch ($packet[0]) {
    if ($packet[0] > 0) {
      bc_handleproto::console_debug("Got game packet type: ".bc_enc::to1hex($packet[0]));
    }
    case 0x00	{ bc_handleproto::console_debug("Got core packet type: ".bc_enc::to1hex($packet[1]));
      switch ($packet[1]) {
	case 0x02	{handle_enc_response(@packet);} # Encryption Request Response
	case 0x03	{handle_reliable_msg(@packet);} # Reliable Message Recieved
	case 0x04	{handle_reliable_ack(@packet);} # Reliable Ack Recieved
	case 0x05	{handle_sync_request(@packet);} # Syncronization Request
	case 0x06	{handle_sync_response(@packet);} # Syncronization Response
	case 0x07	{handle_disconnect();} # The server has told us to "Go Away!"
	case 0x08	{handle_small_chunk(@packet);} # Part of a larger Packet
	case 0x09	{handle_small_chunk_tail(@packet);} # Last part of a Small Chunk
	case 0x0A	{handle_big_chunk(@packet);} # Part of a Big Chunked Packet
	case 0x0E	{handle_cluster(@packet);} # Packet Cluster
	else		{}
      }
    }
    case 0x01	{handle_id_change(@packet);} # Notification of the Bots ID
    case 0x02	{handle_in_game();} # Player is In the Game
    case 0x03	{handle_player_enter(@packet);} # Player(s) entering the Game
    case 0x04	{handle_player_leave(@packet);} # Player Leaving the game
    case 0x05	{handle_large_position(@packet);} # Large position packet
    case 0x06	{handle_player_death(@packet);} # Player death
    case 0x07	{handle_chat_msg(@packet);} # Recieved a Chat Message
    case 0x08	{handle_player_prized(@packet);} # Player prized
    case 0x09	{handle_player_score_update(@packet);} #Player score update
    case 0x0A	{handle_password_response(@packet);} # Recieved Password Response
    case 0x0B	{handle_soccer_goal(@packet);} # Soccer goal
    case 0x0C	{} # Player voice
    case 0x0D	{handle_freq_change(@packet);} # Freq change
    case 0x0E	{handle_turret_link(@packet);} # Turret link
    case 0x0F	{} # Arena Settings Recieved (Comes in four parts)
    case 0x10	{} # File transfer
    case 0x11	{} # Unknown
    case 0x12	{handle_flag_position(@packet);} # Flag position
    case 0x13	{handle_flag_claim(@packet);} # Flag claim
    case 0x14	{handle_flag_victory(@packet);} # Flag victory
    case 0x15	{handle_turret_destroy(@packet);} # Destroy turret link
    case 0x16	{handle_drop_flag(@packet);} # Drop flag
    case 0x17	{} # Unknown
    case 0x18	{handle_synchronization_request(@packet);} # Recieved a Checksum Request
    case 0x19	{} # Request file
    case 0x1A	{handle_score_reset(@packet);} # Reset score(s)
    case 0x1B	{handle_personal_ship_reset();} # Personal ship reset
    case 0x1C	{handle_spec_player(@packet);} # Put player in spectator mode
    case 0x1D	{handle_team_ship_change(@packet);} # Player team and ship changed
    case 0x1E	{handle_banner_change();} # Personal banner changed
    case 0x1F	{} # Player banner changed
    case 0x20	{handle_collected_prize(@packet);} # Collected prize
    case 0x21	{handle_brick_dropped(@packet);} # Brick dropped
    case 0x22	{handle_turf_flag_update(@packet);} # Turf flag update
    case 0x23	{handle_flag_reward(@packet);} # Flag reward granted
    case 0x24	{handle_speed_over(@packet);} # Speed game over
    case 0x25	{handle_ufo_ship(@packet);} # Toggle UFO ship
    case 0x26	{} # Unknown
    case 0x27	{handle_keep_alive(@packet);} # Keep alive
    case 0x28	{handle_small_position(@packet);} # Small position packet
    case 0x29	{handle_map_info(@packet);} # Server Map Checksum
    case 0x2A	{} # Server Map File
    case 0x2B	{handle_set_koth_timer(@packet);} # Set personal koth timer
    case 0x2C	{handle_koth_game_reset(@packet);} # KotH Restarted
    case 0x2D	{} # Unknown
    case 0x2E	{handle_powerball_position(@packet);} # Power-ball position update
    case 0x2F	{handle_arena_list(@packet);} # Arena directory listing
    case 0x30	{} # Got zone banner advertisements
    case 0x31	{handle_goahead();} # Got password, go ahead
    case 0x32	{handle_ship_coord_change(@packet);} # Change personal ship coordinates
    case 0x33	{handle_login_fail(@packet);} # Custom login failure message
    case 0x34	{handle_continuum_version(@packet);} # Continuum version packet
    else	{}
  }
}

# **Server to Client**

#Core packets

# 00 02
sub handle_enc_response {
  my @packet = @_;
  bc_handleproto::console_debug("server enc response: @packet");
  $bc_global::serverKey = bc_enc::dece_int(@packet[2 .. 5]);
  bc_handleproto::console_debug("server key: $bc_global::serverKey");

  $bc_global::initEncryption = 1; #1 for encryption
  $bc_global::state = 1;
  bc_enc::initialize($bc_global::serverKey);

  bc_packetgen::send_password($bc_global::cname, $bc_global::cpass);
  bc_packetgen::send_sync_request();
}

# 00 03
sub handle_reliable_msg {
  my @packet = @_;
  my (@ipacket, $ackID);
  $ackID = bc_enc::dece_int(@packet[2 .. 5]);
  bc_packetgen::send_reliable_ack($ackID);
  @ipacket = @packet[6 .. $#packet];
  $bc_global::slackID[$ackID] = bc_enc::encrypt(@ipacket);
  if (length($bc_global::slackID[$bc_global::ackIDc]) > 0) {
    bc_modules::exec_interpreters(bc_enc::decrypt($bc_global::slackID[$ackID]));
    $bc_global::ackIDc++;
  }
}

sub handle_reliable_ack {
  my @packet = @_;
  my $ackID;
  $ackID = bc_enc::dece_int(@packet[2 .. 5]);
  bc_packetgen::got_ack($ackID);
}

# 00 05
sub handle_sync_request {
  my @packet = @_;
  my $timeStamp;
  bc_handleproto::console_debug("server sync request: @packet");
  $timeStamp = bc_enc::dece_int(@packet[2 .. 5]);
  bc_packetgen::send_sync_response($timeStamp);
}

# 00 06
sub handle_sync_response {
  # <0x06(1)> <Received timestamp(4)> <Timestamp(4)>

  my @packet = @_;
  my ($pingTime, $pongTime, $pingCurrent);
  bc_handleproto::console_debug("server sync response: @packet");
  $pingTime = bc_enc::dece_int(@packet[2 .. 5]);
  $pongTime = bc_enc::dece_int(@packet[6 .. 9]);
  $pingCurrent = int(bc_handleproto::getTicks() / 10) - $pingTime;
}

# 00 07
sub handle_disconnect {
  $bc_global::state = 2;
  print "disconnected\n";
}

# 00 08
sub handle_small_chunk {
  # Offset	Length	Description
  # 0		1		Type Byte 0x08
  # 1		*		Data... *1

  my @packet = @_;
  my (@tempbuf, $i);
  our @buffer;

  undef @tempbuf;
  $i = 0;
  foreach (@packet[2 .. $#packet]) {
    $tempbuf[$i] = $_;
    $i++
  }
  push(@buffer, @tempbuf);

}

# 00 09
sub handle_small_chunk_tail {
  # Offset	Length	Description
  # 0		1		Type Byte 0x09
  # 1		*		Data... *1

  my @packet = @_;
  my (@tempbuf, $i);
  our @buffer;

  undef @tempbuf;
  $i = 0;
  foreach (@packet[2 .. $#packet]) {
    $tempbuf[$i] = $_;
    $i++
  }
  push(@buffer, @tempbuf);
  bc_modules::exec_interpreters(@buffer);
  undef @buffer;
}

# 00 0A
sub handle_big_chunk {
  # Offset	Length	Description
  # 0		1		Type Byte 0x0A
  # 1		4		Total Length
  # 5		*		Data... *1

  my @packet = @_;
  my ($size, $i, @bigBuffer);

  $size = bc_enc::dece_int(@packet[2 .. 5]);
  $i = 0;
  foreach (@packet[6 .. $#packet]) {
    $bigBuffer[$i] = $_;
    if (scalar(@bigBuffer) >= $size) {
      bc_modules::exec_interpreters();
      undef @bigBuffer;
      last;
    }
    $i++;
  }
}

# 00 0E
sub handle_cluster {
  # Offset	Length	Description
  # 0		1		Type Byte 0x0E
  # The following are repeated until the end of the message
  # >1		1		Packet Length
  # >2		*		Packet...

  my @packet = @_;
  my ($size, @temp, $i);

  $i = 2;
  while ($i <= $#packet) {
    $size = $packet[$i];
    if (($#packet - $i + 1) >= $size) {
      @temp = @packet[$i+1 .. $i+$size];
      bc_handleproto::console_debug("cluster: @temp");
      bc_modules::exec_interpreters(@temp);
    }
    $i += 1 + $size;
  }
}

# Game Packets

# 01
sub handle_id_change {
  # Offset	Length	Description
  # 0		1		Type Byte 0x01
  # 1		2		New Player ID *1

  my @packet = @_;
  my $botid;

  $botid = bc_enc::dece_short(@packet[1 .. 2]);
}

# 02
sub handle_in_game {
  bc_packetgen::send_chat_message(2, 0, "*relkills 1");
  bc_handleproto::console_debug("You are now in the game");
}

# 03
sub handle_player_enter {
  # Offset	Length	Description
  # 0		1		Type Byte 0x03
  # 1		1		Ship Type
  # 2		1		? Unknown (0x00)
  # 3		20		Name
  # 23		20		Squad
  # 43		4		Flag Points
  # 47		4		Kill Points
  # 51		2		User ID
  # 53		2		Wins
  # 55		2		Losses
  # 57		6		Audio and Other Stuff

  my @packet = @_;
  my ($playerID, $count, $i);
  our (@p, @ids);

  $count = 0;
  $i = 0;
  while(1) {
    $playerID = bc_enc::dece_short(@packet[$count+51 .. $count+52]);
    unless(grep {$_ == $playerID} @ids) {
      push(@ids, $playerID);
    }

    $p[$playerID]{'ship_type'} = @packet[$count+1];
    $p[$playerID]{'name'} = pack("C*", @packet[$count+3 .. $count+22]);
    $p[$playerID]{'squad'} = pack("C*", @packet[$count+23 .. $count+42]);
    $p[$playerID]{'flag_points'} = bc_enc::dece_int(@packet[$count+43 .. $count+46]);
    $p[$playerID]{'kill_points'} = bc_enc::dece_int(@packet[$count+47 .. $count+50]);
    $p[$playerID]{'user_id'} = bc_enc::dece_short(@packet[$count+51 .. $count+52]);
    $p[$playerID]{'wins'} = bc_enc::dece_short(@packet[$count+53 .. $count+54]);
    $p[$playerID]{'losses'} = bc_enc::dece_short(@packet[$count+55 .. $count+56]);
    $p[$playerID]{'audio'} = join(" ", @packet[$count+57 .. $count+62]);
    bc_handleproto::console_write("Player $p[$playerID]{'name'} (".$playerID.") entered");
    if (@packet[$count+64] != 03) {
      last;
    }
    $count += 64;
    $i++
  }
}

# 04
sub handle_player_leave {
  # Offset	Length	Description
  # 0		1		Type Byte 0x04
  # 1		2		Player ID

  my @packet = @_;
  my ($playerID, $i);
  our (@p, @ids);

  $playerID = bc_enc::dece_short(@packet[1 .. 2]); 
  bc_handleproto::console_write("Player leaving: $playerID - $p[$playerID]{'name'}");

  splice(@p, $playerID, 1, undef);
  for ($i = 0; $i < scalar(@ids); $i++) {
    if ($ids[$i] == $playerID) {
      splice(@ids, $i, 1);
    }
  }
}

# 05
sub handle_large_position {
  # Offset	Length	Description
  # 0		1		Type Byte 0x05
  # 1		1		Direction	(0 ... 360)
  # 2		2		Time Stamp
  # 4		2		X Pixels	(0 ... 16384)
  # 6		2		Y Velocity
  # 8		2		Player ID
  # 10		2		X Velocity
  # 12		1		Checksum
  # 13		1		Toggleables *1
  # 14		1		Ping
  # 15		2		Y Pixels	(0 ... 16384)
  # 17		2		Bounty
  # 19		1		Weapon Parameters
  # 20		1		Weapon Type
  # 
  # 21		2		Energy *2 *3	(Optional)
  # 23		2		S2C Lag *2	(Optional)
  # 25		2		Timer *2	(Optional)
  # 27		4		Item Info *2	(Optional)

  my @packet = @_;
  my $lastel;
  my ($direction, $timeStamp, $XPixels, $YVelocity, $playerID, $XVelocity, $checksum, $toggleables, $ping, $YPixels, $bounty, $weaponParameters, $weaponType, $energy, $lag, $timers, $items);

  $lastel = $#packet;

  $direction = $packet[1];
  $timeStamp = bc_enc::dece_short(@packet[2 .. 3]);
  $XPixels = bc_enc::dece_short(@packet[4 .. 5]);
  $YVelocity = bc_enc::dece_short(@packet[6 .. 7]);
  $playerID = bc_enc::dece_short(@packet[8 .. 9]);
  $XVelocity = bc_enc::dece_short(@packet[10 .. 11]);
  $checksum = $packet[12];
  $toggleables = $packet[13];
  $ping = $packet[14];
  $YPixels = bc_enc::dece_short(@packet[15 .. 16]);
  $bounty = bc_enc::dece_short(@packet[17 .. 18]);
  $weaponParameters = $packet[19];
  $weaponType = $packet[20];
  if ($lastel >= 21) {
    $energy = bc_enc::dece_short(@packet[21 .. 22]);
    $lag = bc_enc::dece_short(@packet[23 .. 24]);
    $timers = bc_enc::dece_short(@packet[25 .. 26]);
    $items = bc_enc::dece_int(@packet[27 .. 30]);
  }
}

# 06
sub handle_player_death {
  # Offset	Length	Description
  # 0		1		Type Byte 0x06
  # 1		1		? Unknown
  # 2		2		Killer ID
  # 4		2		Killed ID
  # 6		2		Bounty *1
  # 8		2		? Unknown

  my @packet = @_;
  my ($killerID, $killedID, $bounty);
  our @p;

  $killerID = bc_enc::dece_short(@packet[2 .. 3]);
  $killedID = bc_enc::dece_short(@packet[4 .. 5]);
  $bounty = bc_enc::dece_short(@packet[6 .. 7]);

  $p[$killerID]{'wins'} = int($p[$killerID]{'wins'} + 1);
  $p[$killedID]{'losses'} = int($p[$killedID]{'losses'} + 1);

  bc_handleproto::console_write("Player $p[$killerID]{'name'} has killed $p[$killedID]{'name'}");
}

# 07
sub handle_chat_msg {

  # Offset	Length	Description
  # 0		1		Type Byte 0x07
  # 1		1		Chat Type *1
  # 2		1		Sound Byte
  # 3		2		Origonator ID *2
  # 5		*		Chat Message

  # *1 - Chat types:
  #	0x00 - Message in green text  [*arena, *zone, ...]
  # 	0x01 - Public macro
  #	0x02 - Public message
  #	0x03 - Team message [// or ']
  #	0x04 - Player to all members of another team  ["Whatever]
  #	0x05 - Private message  [/Whatever]
  #	0x06 - Red warning message [MODERATOR WARNING: Whatever -Whoever]
  #	0x07 - Remote private message	 [(Whoever)> Whatever]
  #	0x08 - Red server errors, without a name tag  (S2C only)
  #	0x09 - Channel message  [;X;Whatever]

  #*2 - Target player's player ID is 0 for public messages and team number for 0x04

  use Switch;
  my @packet = @_;
  my ($player, $playerID, $message, $themsg, $adj);
  our @p;

  $adj = 2;
  switch($packet[1]) {
    case 0x00	{ # Green message
      $playerID = bc_enc::dece_short(@packet[3 .. 4]);
      $themsg = pack("C*", @packet[5 .. $#packet-$adj]);
      $themsg =~ s/[^ -~]//g;
      bc_handleproto::console_write("Arena: $themsg");
    }
    case 0x01	{ # Public macro
      $playerID = bc_enc::dece_short(@packet[3 .. 4]);
      $themsg = pack("C*", @packet[5 .. $#packet-$adj]);
      $themsg =~ s/[^ -~]//g;
      bc_handleproto::console_write("public-$p[$playerID]{'name'}> $themsg");
      bc_command::command(1, $playerID, $themsg);
    }
    case 0x02	{ # Public message
      $playerID = bc_enc::dece_short(@packet[3 .. 4]);
      $themsg = pack("C*", @packet[5 .. $#packet-$adj]);
      $themsg =~ s/[^ -~]//g;
      bc_handleproto::console_write("public-$p[$playerID]{'name'}> $themsg");
      bc_command::command(2, $playerID, $themsg);
    }
    case 0x03	{ # Team message
      $playerID = bc_enc::dece_short(@packet[3 .. 4]);
      $themsg = pack("C*", @packet[5 .. $#packet-$adj]);
      $themsg =~ s/[^ -~]//g;
      bc_handleproto::console_write("team-$p[$playerID]{'name'}> $themsg");
      bc_command::command(3, $playerID, $themsg);
    }
    case 0x04	{ # Enemy team message
      $playerID = bc_enc::dece_short(@packet[3 .. 4]);
      $themsg = pack("C*", @packet[5 .. $#packet-$adj]);
      $themsg =~ s/[^ -~]//g;
      bc_handleproto::console_write("Enemy team-$p[$playerID]{'name'}> $themsg");
    }
    case 0x05	{ # Private message
      $playerID = bc_enc::dece_short(@packet[3 .. 4]);
      $themsg = pack("C*", @packet[5 .. $#packet-$adj]);
      $themsg =~ s/[^ -~]//g;
      bc_handleproto::console_write("private-$p[$playerID]{'name'}> $themsg");
      bc_command::command(5, $playerID, $themsg);
    }
    case 0x06	{ # Moderator warning
      $playerID = bc_enc::dece_short(@packet[3 .. 4]);
      $themsg = pack("C*", @packet[5 .. $#packet-$adj]);
      $themsg =~ s/[^ -~]//g;
      bc_handleproto::console_write("$themsg");
    }
    case 0x07	{ # Remote private message
      $playerID = bc_enc::dece_short(@packet[3 .. 4]);
      $themsg = pack("C*", @packet[5 .. $#packet-$adj]);
      $themsg =~ /\((.+)\)>(.+)/;
      $player = $1;
      $message = $2;
      bc_handleproto::console_write("$themsg");
    }
    case 0x08	{ # Server error [S2C only]
      $themsg = pack("C*", @packet[5 .. $#packet-$adj]);
      $themsg =~ s/[^ -~]//g;
      bc_handleproto::console_write("Server error: $themsg");
    }
    case 0x09	{ # Channel message
      $playerID = bc_enc::dece_short(@packet[3 .. 4]);
      $themsg = pack("C*", @packet[5 .. $#packet-$adj]);
      $themsg =~ s/[^ -~]//g;
      bc_handleproto::console_write("chat-$playerID> $themsg");
    }
    else {
      bc_handleproto::console_debug("Unknown chat message");
    }
  }
}

# 08
sub handle_player_prized {
  # Offset	Length	Description
  # 0		1		Type Byte 0x08
  # 1		4		Timestamp
  # 5		2		X Tiles
  # 7		2		Y Tiles
  # 9		2		Prize
  # 11		2		Player ID

  my @packet = @_;
  my ($timeStamp, $XTiles, $YTiles, $prize, $playerID);

  $timeStamp = bc_enc::dece_int(@packet[1 .. 4]);
  $XTiles = bc_enc::dece_short(@packet[5 .. 6]);
  $YTiles = bc_enc::dece_short(@packet[7 .. 8]);
  $prize = bc_enc::dece_short(@packet[9 .. 10]);
  $playerID = bc_enc::dece_short(@packet[11 .. 12]);

}

# 09
sub handle_player_score_update {
  # Offset	Length	Description
  # 0		1		Type Byte 0x09
  # 1		2		Player ID
  # 3		4		Flag Points
  # 7		4		Kill Points
  # 11		2		Wins
  # 13		2		Losses

  my @packet = @_;
  my ($playerID);
  our @p;

  $playerID = bc_enc::dece_short(@packet[1 .. 2]);
  $p[$playerID]{'flag_points'} = bc_enc::dece_int(@packet[3 .. 6]);
  $p[$playerID]{'kill_points'} = bc_enc::dece_int(@packet[7 .. 10]);
  $p[$playerID]{'wins'} = bc_enc::dece_short(@packet[11 .. 12]);
  $p[$playerID]{'losses'} = bc_enc::dece_short(@packet[13 .. 14]);

}

# 0A
sub handle_password_response {
  # Offset	Length	Description
  # 0		1		Type Byte 0x0A
  # 1		1		Login Response *1
  # 2		4		Server Version *2
  # 6		4		? Unknown
  # 10		4		Subspace.exe Checksum *3
  # 14		4		? Unknown
  # 18		1		? Unknown
  # 19		1		Registration Form Request (Boolean)
  # 20		4		? Unknown
  # 24		4		News.txt Checksum *4
  # 28		8		? Unknown
 
  use Switch;
  my @packet = @_;
  switch ($packet[1]) {
    case 0x00	{ print "Login OK\n"; }
    case 0x01	{ print "Unregistered player\n"; }
    case 0x02	{ print "Wrong password\n"; }
    case 0x03	{ print "Arena is full\n"; }
    case 0x04	{ print "Locked out of zone\n"; }
    case 0x05	{ print "Permission only arena\n"; }
    case 0x06	{ print "Permission to spectate only\n"; }
    case 0x07	{ print "Too many points to play here\n"; }
    case 0x08	{ print "Connection is too slow\n"; }
    case 0x09	{ print "Permission only arena\n"; }
    case 0x0A	{ print "Server is full\n"; }
    case 0x0B	{ print "Invalid name\n"; }
    case 0x0C	{ print "Offensive name\n"; }
    case 0x0D	{ print "No active biller\n"; }
    case 0x0E	{ print "Server busy, try later\n"; }
    case 0x10	{ print "Restricted zone\n"; }
    case 0x11	{ print "Demo version detected\n"; }
    case 0x12	{ print "Too many demo users\n"; }
    case 0x13	{ print "Demo versions not allowed\n"; }
    case 0xFF	{ print "Restricted zone, mod access required\n"; }
  }
  bc_handleproto::console_debug("Server version: ".bc_enc::dece_int(@packet[2 .. 5]));
  if ($packet[19] == 1) {
    bc_handleproto::console_debug("Registration form request (name not registered!)");
  }
  bc_handleproto::console_debug("News checksum: ".bc_enc::dece_int(@packet[24 .. 27]));
  if ($packet[1] == 0x00) {
    $bc_global::state = 1;
    bc_packetgen::send_arena_login(0xFFFF);
  } else {
    $bc_global::state = 2;
    print "disconnected\n";
  }
}

# 0B
sub handle_soccer_goal {
  # Offset	Length	Description
  # 0		1		Type Byte 0x0B
  # 1		2		Score Frequency
  # 3		4		Team Points

  my @packet = @_;
  my ($scoreFreq, $teamPoints);

  $scoreFreq = bc_enc::dece_short(@packet[1 .. 2]);
  $teamPoints = bc_enc::dece_short(@packet[3 .. 6]);
}

# 0D
sub handle_freq_change {
  # Offset	Length	Description
  # 0		1		Type Byte 0x0D
  # 1		2		Player ID
  # 3		2		Frequency

  my @packet = @_;
  my ($playerID, $freq);
  our @p;

  $playerID = bc_enc::dece_short(@packet[1 .. 2]);
  $freq = bc_enc::dece_short(@packet[3 .. 4]);
  $p[$playerID]{'freq'} = $freq;
}

# 0E
sub handle_turret_link {
  # Offset	Length	Description
  # 0		1		Type Byte 0x0E
  # 1		2		Turreter Requester Player ID
  # 3		2		Turreter Destination Player ID *1

  my @packet = @_;
  my ($turretReq, $turretDes);

  $turretReq = bc_enc::dece_short(@packet[1 .. 2]);
  $turretDes = bc_enc::dece_short(@packet[3 .. 4]);
}

# 12
sub handle_flag_position {
  # Offset	Length	Description
  # 0		1		Type Byte 0x12
  # 1		2		Flag ID
  # 3		2		X Tiles
  # 5		2		Y Tiles
  # 7		2		Owner *1

  my @packet = @_;
  my ($flagID, $XTiles, $YTiles, $ownerTeam);

  $flagID = bc_enc::dece_short(@packet[1 .. 2]);
  $XTiles = bc_enc::dece_short(@packet[3 .. 4]);
  $YTiles = bc_enc::dece_short(@packet[5 .. 6]);
  $ownerTeam = bc_enc::dece_short(@packet[7 .. 8]);
}

# 13
sub handle_flag_claim {
  # Offset	Length	Description
  # 0		1		Type Byte 0x13
  # 1		2		Flag ID
  # 3		2		Player ID

  my @packet = @_;
  my ($flagID, $playerID);

  $flagID = bc_enc::dece_short(@packet[1 .. 2]);
  $playerID = bc_enc::dece_short(@packet[3 .. 4]);
}

# 14
sub handle_flag_victory {
  # Offset	Length	Description
  # 0		1		Type Byte 0x14
  # 1		2		Team
  # 3		4		Jackpot Points

  my @packet = @_;
  my ($team, $jackpotPoints);

  $team = bc_enc::dece_short(@packet[1 .. 2]);
  $jackpotPoints = bc_enc::dece_int(@packet[3 .. 6]);
}

# 15
sub handle_turret_destroy {
  # Offset	Length	Description
  # 0		1		Type Byte 0x15
  # 1		2		Turret Driver Player ID *1

  my @packet = @_;
  my $turretDriverID;

  $turretDriverID = bc_enc::dece_short(@packet[1 .. 2]);
}

# 16
sub handle_drop_flag {
  # Offset	Length	Description
  # 0		1		Type Byte 0x16
  # 1		2		Carrier Player ID

  my @packet = @_;
  my $carrierID;

  $carrierID = bc_enc::dece_short(@packet[1 .. 2]);
}

# 18
sub handle_synchronization_request {
  # Offset	Length	Description
  # 0		1		Type Byte 0x18
  # 1		4		Prize Seed Value *1
  # 5		4		Door Seed Value *1
  # 9		4		Timestamp
  # 13		4		Checksum Key *2

  my @packet = @_;
  my ($prizeSeed, $doorSeed, $timeStamp, $checksumKey);

  $prizeSeed = bc_enc::dece_int(@packet[1 .. 4]);
  $doorSeed = bc_enc::dece_int(@packet[5 .. 8]);
  $timeStamp = bc_enc::dece_int(@packet[9 .. 12]);
  $checksumKey = bc_enc::dece_int(@packet[13 .. 16]);
}

# 1A
sub handle_score_reset {
  # Offset	Length	Description
  # 0		1		Type Byte 0x1A
  # 1		2		Player ID *1

  my @packet = @_;
  my $playerID;
  our @p;

  $playerID = bc_enc::dece_short(@packet[1 .. 2]);
  if ($playerID == 0xFFFF) {
    foreach (@p) {
      if (length($_->{'name'}) > 0) {
	$_->{'wins'} = 0;
	$_->{'losses'} = 0;
	$_->{'flag_points'} = 0;
	$_->{'kill_points'} = 0;
      }
    }
  }

}

# 1B
sub handle_personal_ship_reset {
  # Offset	Length	Description
  # 0		1		Type Byte 0x1B 

}

# 1C
sub handle_spec_player {
  # Offset	Length	Description
  # 0		1		Type Byte 0x1C
  # One of the following: *1
  # 1 <-		2		Player ID *2
  # 1 <-		1		Bool: Request Extra Pos Info

  my @packet = @_;
  my ($playerID, $extraPosReq);

  if (scalar(@packet) == 3) {
    $playerID = bc_enc::dece_short(@packet[1 .. 2]);
  } else {
    $extraPosReq = @packet[1];
  }
}

# 1D
sub handle_team_ship_change {
  # Offset	Length	Description
  # 0		1		Type Byte 0x1D
  # 1		1		Ship Type
  # 2		2		Player ID
  # 4		2		Frequency

  my @packet = @_;
  my ($playerID, $freq);
  our @p;

  $playerID = bc_enc::dece_short(@packet[2 .. 3]);
  $p[$playerID]{'ship_type'} = @packet[1];
  $freq = bc_enc::dece_short(@packet[4 .. 5]);
  $p[$playerID]{'freq'} = $freq;
}

# 1E
sub handle_banner_change {
  # Offset	Length	Description
  # 0		1		Type Byte 0x1E
  # 1		1		Bool: ?Unknown? (0x00)

}

# 20
sub handle_collected_prize {
  # Offset	Length	Description
  # 0		1		Type Byte 0x20
  # 1		4		Timestamp
  # 5		2		X Tiles
  # 7		2		Y Tiles
  # 9		2		Prize Type *1
  # 11		2		Player ID

  my @packet = @_;
  my ($timeStamp, $XTiles, $YTiles, $prizeType, $playerID);

  $timeStamp = bc_enc::dece_int(@packet[1 .. 4]);
  $XTiles = bc_enc::dece_short(@packet[5 .. 6]);
  $YTiles = bc_enc::dece_short(@packet[7 .. 8]);
  $prizeType = bc_enc::dece_short(@packet[9 .. 10]);
  $playerID = bc_enc::dece_short(@packet[11 .. 12]);
}

# 21
sub handle_brick_dropped {
  # Offset	Length	Description
  # 0		1		Type Byte 0x21
  # The following are repeated until the end of the message
  # >1		2		X1 Tiles
  # >3		2		Y1 Tiles
  # >5		2		X2 Tiles
  # >7		2		Y2 Tiles
  # >9		2		Team
  # >11		2		Brick ID (sent more than once)
  # >13		4		Timestamp

  my @packet = @_;
  my ($size, $lastel);
  my ($X1Tile, $Y1Tile, $X2Tile, $Y2Tile, $team, $brickID, $timeStamp);

  $size = 0;
  $lastel = $#packet;
  while($lastel >= $size) {
    $X1Tile = bc_enc::dece_short(@packet[$size+1 .. $size+2]);
    $Y1Tile = bc_enc::dece_short(@packet[$size+3 .. $size+4]);
    $X2Tile = bc_enc::dece_short(@packet[$size+5 .. $size+6]);
    $Y2Tile = bc_enc::dece_short(@packet[$size+7 .. $size+8]);
    $team = bc_enc::dece_short(@packet[$size+9 .. $size+10]);
    $brickID = bc_enc::dece_short(@packet[$size+11 .. $size+12]);
    $timeStamp = bc_enc::dece_int(@packet[$size+13 .. $size+16]);
    $size += 16;
  }
}

# 22
sub handle_turf_flag_update {
  # Offset	Length	Description
  # 0		1		Type Byte 0x22
  # The following are repeated until the end of the message
  # >1		2		Team for Flag X

  my @packet = @_;
  my $size;
  my $team;

  $size = 0;
  while($#packet >= $size) {
    $team = bc_enc::dece_short(@packet[$size+1 .. $size+2]);
    $size += 2;
  }
}

# 23
sub handle_flag_reward {
  # 0		1		Type Byte 0x23
  # 1		2		Team
  # 3		2		Points Awarded
  # The following are repeated until the end of the message
  # >5		2		Team *1		(Optional)
  # >7		2		Points Awarded *1	(Optional)

  my @packet = @_;
  my ($size, $lastel);
  my ($team, $pointsAwarded);

  $size = 0;
  $lastel = $#packet;
  while($lastel >= $size) {
    $team = bc_enc::dece_short(@packet[$size+1 .. $size+2]);
    $pointsAwarded = bc_enc::dece_short(@packet[$size+3 .. $size+4]);
    $size += 4;
  }
}

# 24
sub handle_speed_over {
  # Offset	Length	Description
  # 0		1		Type Byte 0x24
  # 1		1		Best
  # 2		2		Your Rank
  # 4		4		Your Score
  # 8		4		Player 1 Score
  # 12		4		Player 2 Score
  # 16		4		Player 3 Score
  # 20		4		Player 4 Score
  # 24		4		Player 5 Score
  # 28		2		Player 1 ID
  # 30		2		Player 2 ID
  # 32		2		Player 3 ID
  # 34		2		Player 4 ID
  # 36		2		Player 5 ID

  my @packet = @_;
  my ($best, $myRank, $myScore, $p1Score, $p2Score, $p3Score, $p4Score, $p5Score, $p1ID, $p2ID, $p3ID, $p4ID, $p5ID);

  $best = @packet[1];
  $myRank = bc_enc::dece_short(@packet[2 .. 3]);
  $myScore = bc_enc::dece_int(@packet[4 .. 7]);
  $p1Score = bc_enc::dece_int(@packet[8 .. 11]);
  $p2Score = bc_enc::dece_int(@packet[12 .. 15]);
  $p3Score = bc_enc::dece_int(@packet[16 .. 19]);
  $p4Score = bc_enc::dece_int(@packet[20 .. 23]);
  $p5Score = bc_enc::dece_int(@packet[24 .. 27]);
  $p1ID = bc_enc::dece_short(@packet[28 .. 29]);
  $p2ID = bc_enc::dece_short(@packet[30 .. 31]);
  $p3ID = bc_enc::dece_short(@packet[32 .. 33]);
  $p4ID = bc_enc::dece_short(@packet[34 .. 35]);
  $p5ID = bc_enc::dece_short(@packet[36 .. 37]);
}

# 25
sub handle_ufo_ship {
  # Offset	Length	Description
  # 0		1		Type Byte 0x24
  # 1		1		Boolean: Available?

  my @packet = @_;
  my $available;

  $available = $packet[1];
}

# 27
sub handle_keep_alive {
  # Offset	Length	Description
  # 0		1		Type Byte 0x27
}

# 28
sub handle_small_position {
  # Offset	Length	Description
  # 0		1		Type Byte 0x28
  # 1		1		Direction	(0 ... 360)
  # 2		2		Timestamp
  # 4		2		X Pixels	(0 ... 16384)
  # 6		1		Ping
  # 7		1		Bounty
  # 8		1		Player ID *1
  # 9		1		Toggleables *2
  # 10		2		Y Velocity *3
  # 12		2		Y Pixels	(0 ... 16384)
  # 14		2		X Velocity *3
  #
  # 16		2		Energy *4 *5	(Optional) 
  # 18		2		S2C Lag *4	(Optional)
  # 20		2		Timers *4	(Optional)
  # 22		4		Items *4 *6	(Optional)

  my @packet = @_;
  my $lastel;
  my ($direction, $timeStamp, $XPixels, $ping, $bounty, $playerID, $toggleables, $YVelocity, $YPixels, $XVelocity, $energy, $lag, $timers, $items);

  $lastel = $#packet;

  $direction = $packet[1];
  $timeStamp = bc_enc::dece_short(@packet[2 .. 3]);
  $XPixels = bc_enc::dece_short(@packet[4 .. 5]);
  $ping = $packet[6];
  $bounty = $packet[7];
  $playerID = $packet[8];
  $toggleables = $packet[9];
  $YVelocity = bc_enc::dece_short(@packet[10 .. 11]);
  $YPixels = bc_enc::dece_short(@packet[12 .. 13]);
  $XVelocity = bc_enc::dece_short(@packet[14 .. 15]);
  if ($lastel >= 16) {
    $energy = bc_enc::dece_short(@packet[16 .. 17]);
    $lag = bc_enc::dece_short(@packet[18 .. 19]);
    $timers = bc_enc::dece_short(@packet[20 .. 21]);
    $items = bc_enc::dece_int(@packet[22 .. 25]);
  }
}

# 29
sub handle_map_info {
  # Offset	Length	Description
  # 0		1		Type Byte 0x29
  # 1		16		Map File Name
  # 17		4		Map Checksum *1

  my @packet = @_;
  my ($mapFile, $mapChecksum);

  $mapFile = pack("C*", @packet[1 .. 16]);
  $mapChecksum = bc_enc::dece_int(@packet[17 .. 20]);
}

# 2B
sub handle_set_koth_timer {
  # Offset	Length	Description
  # 0		1		Type Byte 0x2B
  # 1		4		New Timer Value

  my @packet = @_;
  my $timer;

  $timer = bc_enc::dece_int(@packet[1 .. 4]);
}

# 2C
sub handle_koth_game_reset {
  # Offset	Length	Description
  # 0		1		Type Byte 0x2C
  # 1		1		Adding KoTH Timer
  # 2		4		Timer Value
  # 6		2		Player ID

  my @packet = @_;
  my ($addingKothTimer, $timerValue, $playerID);

  $addingKothTimer = $packet[1];
  $timerValue = bc_enc::dece_int(@packet[2 .. 5]);
  $playerID = bc_enc::dece_short(@packet[6 .. 7]);
}

# 2E
sub handle_powerball_position {
  # Offset	Length	Description
  # 0		1		Type Byte 0x2E
  # 1		1		Powerball ID
  # 2		2		X Pixels
  # 4		2		Y Pixels
  # 6		2		X Velocity
  # 8		2		Y Velocity
  # 10		2		Owner ID
  # 12		4		Timestamp

  my @packet = @_;
  my ($powerballID, $XPixels, $YPixels, $XVelocity, $YVelocity, $ownerID, $timeStamp);

  $powerballID = $packet[1];
  $XPixels = bc_enc::dece_short(@packet[2 .. 3]);
  $YPixels = bc_enc::dece_short(@packet[4 .. 5]);
  $XVelocity = bc_enc::dece_short(@packet[6 .. 7]);
  $YVelocity = bc_enc::dece_short(@packet[8 .. 9]);
  $ownerID = bc_enc::dece_short(@packet[10 .. 11]);
  $timeStamp = bc_enc::dece_int(@packet[12 .. 15]);
} 

# 2F
sub handle_arena_list {
  # Offset	Length	Description
  # 0		1		Type Byte 0x2F
  # The following are repeated until the end of the message
  # >1		*		Arena Name (Null Terminated) *1
  # >?		2		Arena Population

  my @packet = @_;
  my ($size, $lastel, $termLoc);
  my ($arenaName, $arenaPopulation);

  $lastel = $#packet;
  $arenaName = "";
  $size = 1;
  while ($lastel >= $size) {
    foreach (@packet[$size .. $lastel]) {
      if ($_ != 0) {
	$arenaName = $arenaName.pack("C", $_);
      } else {
	$termLoc = length($arenaName) + $size;
	$arenaPopulation = bc_enc::dece_short(@packet[$termLoc+1 .. $termLoc+2]);
	# $arenaName, $arenaPopulation is here
	bc_handleproto::console_debug("$arenaName - $arenaPopulation\n");
	last;
      }
    }
    $arenaName = "";
    $size = $termLoc + 3;
  }
}

# 30
sub handle_banner_advertisement {
  # Offset	Length	Description
  # 0		1		Type Byte 0x30
  # 1		1		Display Mode
  # 2		2		Width
  # 4		2		Height
  # 6		4		Duration
  # 10		*		Banner *1

  my @packet = @_;
  my ($displayMode, $width, $height, $duration, $banner);

  # Not impletmented
}

# 31
sub handle_goahead {
  #bc_packetgen::send_arena_login(0);
  #bc_packetgen::send_chat_message(02, 0x00, "This is a test");
}

# Continuum

# 32
sub handle_ship_coord_change {
  # Offset	Length	Description
  # 0		1		Type Byte 0x32
  # 1		2		X Tiles
  # 3		2		Y Tiles

  my @packet = @_;
  my ($XTiles, $YTiles);

  $XTiles = bc_enc::dece_short(@packet[1 .. 2]);
  $YTiles = bc_enc::dece_short(@packet[3 .. 4]);
}

# 33
sub handle_login_fail {
  # Offset	Length	Description
  # 0		1		Type Byte 0x33
  # 1		?		Message *1

  my @packet = @_;
  my $message;

  $message = "";
  foreach (@packet[1 .. $#packet]) {
    if ($_ != 0) {
      $message = $message.pack("C", $_);
    } else {
      last;
    }
  }
}

# 34
sub handle_continuum_version {
  # Offset	Length	Description
  # 0		1		Type Byte 0x34
  # 1		2		Version number
  # 3		4		Checksum

  my @packet = $_;
  my ($version, $checksum);

  $version = bc_enc::dece_short(@packet[1 .. 2]);
  $checksum = bc_enc::dece_int(@packet[3 .. 6]);
}

# Signal EOF
1;
